import functools
from collections import OrderedDict as odict
from typing import Any, Callable, List, Optional

from valkit import add_validator_magic
from valkit.common import valid_bool, valid_number, valid_string_list

from . import platform, syntax


# =====
@functools.lru_cache()
def compile_acl_text(text, vendor, allow_ignore=False):
    return _compile_acl(
        trees=[syntax.parse_text(text, _PARAMS_SCHEME)],
        reverse_prefix=platform.VENDOR_REVERSES[vendor],
        allow_ignore=allow_ignore,
        vendor=vendor,
    )


@functools.lru_cache()
def compile_ref_acl_text(text):
    return _compile_acl(
        trees=[syntax.parse_text(text, _PARAMS_SCHEME)],
        reverse_prefix="",
        allow_ignore=False,
    )


@add_validator_magic
def valid_bool_list(
    arg: Any,
    delim: str = r"[,\t ]+",
    subval: Optional[Callable[[Any], Any]] = None,
    strip: bool = False,
) -> List[bool]:
    arg = valid_string_list(arg, delim, subval, strip)
    arg = [valid_bool(x, strip) for x in arg]
    return arg


# =====
_PARAMS_SCHEME = {
    "global": {
        "validator": valid_bool,
        "default":   False,
        "uniter":    (lambda a, b: a or b),
    },
    "cant_delete": {
        "validator": valid_bool_list,
        "default":   (lambda raw_rule: [raw_rule.startswith("interface")]),  # FIXME: ужас какой
        "uniter": (lambda a, b: a + b)
    },
    "prio": {
        "validator": (lambda s: valid_number(s, min=0, type=int)),
        "default":   0,
        "uniter":    max,
    },
    "generator_names": {
        "validator": valid_string_list,
        "default": [],
        "uniter": (lambda a, b: a + b)
    }
}


# =====
def _compile_acl(trees, reverse_prefix, allow_ignore=False, vendor=""):
    rules = {"local": odict(), "global": odict()}
    for (rule_id, attrs) in _merge_toplevel(trees).items():
        if attrs["type"] == "ignore" and not allow_ignore:
            raise NotImplementedError("ACL does not support ignore-rules")
        rule = {
            "type": attrs["type"],
            "attrs": {
                "direct_regexp": syntax.compile_row_regexp(attrs["row"]),
                "reverse_regexp": syntax.compile_row_regexp(_make_reverse(attrs["row"], reverse_prefix)),
                "cant_delete": attrs["params"]["cant_delete"],
                "prio": attrs["params"]["prio"],
                "generator_names": attrs["params"]["generator_names"],
                "vendor": vendor,
                "context": attrs["context"],
            },
            "children": None,
        }
        if not attrs["params"]["global"] and not attrs["type"] == "ignore":
            rule["children"] = _compile_acl(attrs["children"], reverse_prefix, allow_ignore, vendor)
        rules["global" if attrs["params"]["global"] else "local"][rule_id] = rule
    return rules


def _merge_toplevel(trees):
    merged = odict()
    for tree in trees:
        for attrs in tree.values():
            rule_id = ("!" if attrs["type"] == "ignore" else "") + attrs["row"]
            if rule_id not in merged:
                merged[rule_id] = attrs
                merged[rule_id]["children"] = [attrs["children"]] if attrs["children"] else []
                continue

            for (key, value) in attrs["params"].items():
                if key in merged[rule_id]["params"]:
                    uniter = _PARAMS_SCHEME[key]["uniter"]
                    merged[rule_id]["params"][key] = uniter(merged[rule_id]["params"][key], value)
                else:
                    merged[rule_id]["params"][key] = value

            if attrs["children"]:
                merged[rule_id]["children"].append(attrs["children"])
    return merged


@functools.lru_cache()
def _make_reverse(row, reverse_prefix):
    if row.startswith(reverse_prefix + " "):
        return row[len(reverse_prefix + " "):]
    else:
        return "%s %s" % (reverse_prefix, row)



import argparse
import itertools
import json
import operator
import os
import platform
import subprocess
import shutil
import sys
from contextlib import ExitStack, contextmanager
from typing import Optional, Tuple, Iterable

import tabulate
import yaml
from contextlog import get_logger
from valkit.python import valid_logging_level

from annet.deploy import get_fetcher, get_deployer
from annet import api, cli_args, filtering, generators
from annet.api import collapse_texts, Deployer
from annet.argparse import ArgParser, subcommand
from annet.diff import gen_sort_diff
from annet.gen import Loader, old_raw
from annet.lib import get_context_path, repair_context_file
from annet.output import output_driver_connector, OutputDriver
from annet.storage import get_storage, Device


def fill_base_args(parser: ArgParser, pkg_name: str, logging_config: str):
    parser.add_argument("--log-level", default="WARN", type=valid_logging_level,
                        help="Уровень детализации логов (DEBUG, DEBUG2 (with comocutor debug), INFO, WARN, CRITICAL)")
    parser.add_argument("--pkg_name", default=pkg_name, help=argparse.SUPPRESS)
    parser.add_argument("--logging_config", default=logging_config, help=argparse.SUPPRESS)


def list_subcommands():
    return globals().copy()


def _gen_current_items(
        config,
        stdin,
        loader: Loader,
        output_driver: OutputDriver,
        gen_args: cli_args.GenOptions,
) -> Iterable[Tuple[str, str, bool]]:
    for device, result in old_raw(
        args=gen_args,
        loader=loader,
        config=config,
        stdin=stdin,
        do_files_download=True,
        use_mesh=False,
    ):
        if device.hw.vendor != "pc":
            destname = output_driver.cfg_file_names(device)[0]
            yield (destname, result, False)
        else:
            for entire_path, entire_data in sorted(result.items(), key=operator.itemgetter(0)):
                if entire_data is None:
                    entire_data = ""
                destname = output_driver.entire_config_dest_path(device, entire_path)
                yield (destname, entire_data, False)


@contextmanager
def get_loader(gen_args: cli_args.GenOptions, args: cli_args.QueryOptionsBase):
    exit_stack = ExitStack()
    storages = []
    with exit_stack:
        connector, conf_params = get_storage()
        storage_opts = connector.opts().parse_params(conf_params, args)
        storages.append(exit_stack.enter_context(connector.storage()(storage_opts)))
        yield Loader(*storages, args=gen_args, no_empty_warning=args.query.is_empty())


@subcommand(is_group=True)
def show():
    """ A group of commands for showing parameters/configurations/data from deivces and data sources """
    pass


@subcommand(cli_args.QueryOptions, cli_args.opt_config, cli_args.FileOutOptions, parent=show)
def show_current(args: cli_args.QueryOptions, config, arg_out: cli_args.FileOutOptions) -> None:
    """ Show current devices' configuration """
    gen_args = cli_args.GenOptions(args, no_acl=True)
    output_driver = output_driver_connector.get()
    with get_loader(gen_args, args) as loader:
        if not loader.devices:
            get_logger().error("No devices found for %s", args.query)

        items = _gen_current_items(
            loader=loader,
            output_driver=output_driver,
            gen_args=gen_args,
            stdin=args.stdin(config=config),
            config=config,
        )
        output_driver.write_output(arg_out, items, len(loader.devices))


@subcommand(cli_args.QueryOptions, cli_args.FileOutOptions, parent=show)
def show_device_dump(args: cli_args.QueryOptions, arg_out: cli_args.FileOutOptions):
    """ Show a dump of network devices' structure """
    def _show_device_dump_items(devices):
        for device in devices:
            get_logger(host=device.hostname)  # add hostname into context
            if hasattr(device, "dump"):
                yield (
                    device.hostname,
                    "\n".join(device.dump("device")),
                    False,
                )
            else:
                get_logger().warning("method `dump` not implemented for %s", type(device))
    arg_gens = cli_args.GenOptions(arg_out, args)
    with get_loader(arg_gens, args) as loader:
        if not loader.devices:
            get_logger().error("No devices found for %s", args.query)
        output_driver_connector.get().write_output(
            arg_out,
            _show_device_dump_items(loader.devices),
            len(loader.device_ids),
        )


@subcommand(cli_args.ShowGeneratorsOptions, parent=show)
def show_generators(args: cli_args.ShowGeneratorsOptions):
    """ List applicable generators (for a device if query is set) """
    arg_gens = cli_args.GenOptions(args)
    with get_loader(arg_gens, args) as loader:
        device: Optional[Device] = None
        devices = loader.devices
        if len(devices) == 1:
            device = devices[0]
        elif len(devices) > 1:
            get_logger().error("cannot show generators for more than one device at once")
            sys.exit(1)
        elif len(devices) == 0 and not args.query.is_empty():
            # the error message will be logged in get_loader()
            sys.exit(1)

        if not device:
            found_generators = loader.iter_all_gens()
        else:
            found_generators = []
            gens = loader.resolve_gens(loader.devices)
            for g in gens.partial[device]:
                acl_func = g.acl_safe if args.acl_safe else g.acl
                if g.supports_device(device) and acl_func(device):
                    found_generators.append(g)
            for g in gens.entire[device]:
                if g.supports_device(device) and g.path(device):
                    found_generators.append(g)
            for g in gens.json_fragment[device]:
                if g.supports_device(device) and g.path(device) and g.acl(device):
                    found_generators.append(g)

    output_data = []
    for g in found_generators:
        output_data.append({
            "name": g.__class__.__name__,
            "type": g.TYPE,
            "tags": g.TAGS,
            "module": g.__class__.__module__,
            "description": generators.get_description(g.__class__),
        })

    if args.format == "json":
        print(json.dumps(output_data))

    elif args.format == "text":
        keyfunc = operator.itemgetter("type")
        for gen_type, gens in itertools.groupby(sorted(output_data, key=keyfunc, reverse=True), keyfunc):
            print(tabulate.tabulate(
                [(g["name"], ", ".join(g["tags"]), g["module"],  g["description"]) for g in gens],
                [f"{gen_type}-Class", "Tags", "Module", "Description"],
                tablefmt="orgtbl",
            ))
            print()


@subcommand(cli_args.ShowGenOptions)
def gen(args: cli_args.ShowGenOptions):
    """ Generate configuration for devices """
    with get_loader(args, args) as loader:
        (success, fail) = api.gen(args, loader)

        out = [item for items in success.values() for item in items]
        output_driver = output_driver_connector.get()
        if args.dest is None:
            text_mapping = {item[0]: item[1] for item in out}
            out = [(",".join(key), value, False) for key, value in collapse_texts(text_mapping).items()]

        out.extend(output_driver.format_fails(fail, loader.device_fqdns))
        total = len(success) + len(fail)
        if not total:
            get_logger().error("No devices found for %s", args.query)
        output_driver.write_output(args, out, total)


@subcommand(cli_args.ShowDiffOptions)
def diff(args: cli_args.ShowDiffOptions):
    """ Generate configuration for devices and show a diff with current configuration using the rulebook """
    with get_loader(args, args) as loader:
        filterer = filtering.filterer_connector.get()
        device_ids = loader.device_ids
        output_driver_connector.get().write_output(
            args,
            gen_sort_diff(api.diff(args, loader, device_ids, filterer), args),
            len(loader.device_ids)
        )


@subcommand(cli_args.ShowPatchOptions)
def patch(args: cli_args.ShowPatchOptions):
    """ Generate configuration patch for devices """
    with get_loader(args, args) as loader:
        (success, fail) = api.patch(args, loader)

        out = [item for items in success.values() for item in items]
        output_driver = output_driver_connector.get()
        out.extend(output_driver.format_fails(fail, loader.device_fqdns))
        total = len(success) + len(fail)
        if not total:
            get_logger().error("No devices found for %s", args.query)
        output_driver.write_output(args, out, total)


@subcommand(cli_args.DeployOptions)
def deploy(args: cli_args.DeployOptions):
    """ Generate and deploy configuration for devices """

    deployer = Deployer(args)
    filterer = filtering.filterer_connector.get()
    fetcher = get_fetcher()
    deploy_driver = get_deployer()

    with get_loader(args, args) as loader:
        return api.deploy(
            args=args, loader=loader, deployer=deployer,
            deploy_driver=deploy_driver, filterer=filterer,
            fetcher=fetcher,
        )


@subcommand(cli_args.FileDiffOptions)
def file_diff(args: cli_args.FileDiffOptions):
    """ Generate a diff between files or directories using the rulebook """
    (success, fail) = api.file_diff(args)
    out = []
    output_driver = output_driver_connector.get()
    if not args.fails_only:
        out.extend(item for items in success.values() for item in items)
    out.extend(output_driver.format_fails(fail))
    # todo отрефакторить логику с отображением хоста в диффе: передавать в write_output явно критерий
    output_driver.write_output(args, out, len(out) + 1)


@subcommand(cli_args.FilePatchOptions)
def file_patch(args: cli_args.FilePatchOptions):
    """ Generate configuration patch for files or directories """
    (success, fail) = api.file_patch(args)
    out = []
    output_driver = output_driver_connector.get()
    if not args.fails_only:
        out.extend(item for items in success.values() for item in items)
    out.extend(output_driver.format_fails(fail))
    output_driver.write_output(args, out, len(out))


@subcommand(is_group=True)
def context():
    """ A group of commands for manipulating context.

    By default, the context file is located in '~/.annet/context.yml',
    but it can be set with the ANN_CONTEXT_CONFIG_PATH environment variable.
    """
    context_touch()


@subcommand(parent=context)
def context_touch():
    """ Show the context file path, and if the file is not present, create it with the default configuration """
    print(get_context_path(touch=True))


@subcommand(cli_args.SelectContext, parent=context)
def context_set_context(args: cli_args.SelectContext):
    """ Set the current active context.

    The selected context is used by default unless the environment variable ANN_SELECTED_CONTEXT is set
    """
    with open(path := get_context_path(touch=True)) as f:
        data = yaml.safe_load(f)
    if args.context_name not in data.get("context", {}):
        raise KeyError(f"Cannot select context with name '{args.context_name}'. "
                       f"Available options are: {list(data.get('context', []))}")
    data["selected_context"] = args.context_name
    with open(path, "w") as f:
        yaml.dump(data, f, sort_keys=False)


@subcommand(parent=context)
def context_edit():
    """ Open the context file using an editor from the EDITOR environment variable.

    If the EDITOR variable is not set, default variables are: "notepad.exe" for Windows and "vi" otherwise
    """
    if e := os.getenv("EDITOR"):
        editor = e
    elif platform.system() == "Windows":
        editor = "notepad.exe"
    elif shutil.which("vim"):
        editor = "vim"
    else:
        editor = "vi"
    path = get_context_path(touch=True)
    proc = subprocess.Popen([editor, path])
    proc.wait()


@subcommand(parent=context)
def context_repair():
    """ Try to fix the context file's structure if it was generated for the older versions of annet """
    repair_context_file()



import abc
from typing import Type

from annet.connectors import Connector


class _FiltererConnector(Connector["Filterer"]):
    name = "Filterer"
    ep_name = "filterer"
    ep_by_group_only = "annet.connectors.filterer"

    def _get_default(self) -> Type["Filterer"]:
        return NopFilterer


filterer_connector = _FiltererConnector()


class Filterer(abc.ABC):
    @abc.abstractmethod
    def for_ifaces(self, device, ifnames) -> str:
        pass

    @abc.abstractmethod
    def for_peers(self, device, peers_allowed) -> str:
        pass

    @abc.abstractmethod
    def for_policies(self, device, policies_allowed) -> str:
        pass


class NopFilterer(Filterer):
    def for_ifaces(self, device, ifnames) -> str:
        return ""

    def for_peers(self, device, peers_allowed) -> str:
        return ""

    def for_policies(self, device, policies_allowed) -> str:
        return ""



from __future__ import annotations

import dataclasses
import importlib
import os
import textwrap
from collections import OrderedDict as odict
from typing import (
    FrozenSet,
    Iterable,
    List,
    Optional,
    Union,
)

from annet.annlib.rbparser.acl import compile_acl_text
from contextlog import get_logger

from annet.storage import Device

from annet import patching, tabparser, tracing
from annet.cli_args import GenSelectOptions, ShowGeneratorsOptions
from annet.lib import (
    get_context,
)
from annet.tracing import tracing_connector
from annet.types import (
    GeneratorEntireResult,
    GeneratorJSONFragmentResult,
    GeneratorPartialResult,
    GeneratorPartialRunArgs,
    GeneratorResult,
)
from .base import (
    BaseGenerator,
    TextGenerator as TextGenerator,
    ParamsList as ParamsList,
)
from .exceptions import NotSupportedDevice, GeneratorError
from .jsonfragment import JSONFragment
from .partial import PartialGenerator
from .entire import Entire
from .ref import RefGenerator
from .perf import GeneratorPerfMesurer
from .result import RunGeneratorResult

# =====
DISABLED_TAG = "disable"


# =====
def get_list(args: ShowGeneratorsOptions):
    if args.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = args.generators_context
    return {
        cls.__class__.__name__: {
            "type": cls.TYPE,
            "tags": set(cls.TAGS),
            "description": get_description(cls.__class__),
        }
        for cls in _get_generators(get_context()["generators"], None)
    }


def get_description(gen_cls) -> str:
    return textwrap.dedent(" ".join([
        (gen_cls.__doc__ or ""),
        ("Disabled. Use '-g %s' to enable" % gen_cls.__name__ if DISABLED_TAG in gen_cls.TAGS else "")
    ])).strip()


def validate_genselect(gens: GenSelectOptions, all_classes):
    logger = get_logger()
    unknown_err = "Unknown generator alias %s"
    all_aliases = {
        alias
        for cls in all_classes
        for alias in cls.get_aliases()
    }
    for gen_set in (gens.allowed_gens, gens.force_enabled):
        for alias in set(gen_set or ()) - all_aliases:
            logger.error(unknown_err, alias)
            raise Exception(unknown_err % alias)


@dataclasses.dataclass
class Generators:
    """Collection of various types of generators."""

    partial: List[PartialGenerator] = dataclasses.field(default_factory=list)
    entire: List[Entire] = dataclasses.field(default_factory=list)
    ref: List[RefGenerator] = dataclasses.field(default_factory=list)
    json_fragment: List[JSONFragment] = dataclasses.field(default_factory=list)


def build_generators(storage, gens: GenSelectOptions, device: Optional[Device] = None) -> Generators:
    """Return generators that meet the gens filter conditions."""
    if gens.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = gens.generators_context
    all_generators = _get_generators(get_context()["generators"], storage, device)
    ref_generators = _get_ref_generators(get_context()["generators"], storage, device)
    validate_genselect(gens, all_generators)
    classes = list(select_generators(gens, all_generators))
    partial = [obj for obj in classes if obj.TYPE == "PARTIAL"]
    entire = [obj for obj in classes if obj.TYPE == "ENTIRE"]
    entire = list(sorted(entire, key=lambda x: x.prio, reverse=True))
    json_fragment = [obj for obj in classes if obj.TYPE == "JSON_FRAGMENT"]
    return Generators(
        partial=partial,
        entire=entire,
        json_fragment=json_fragment,
        ref=ref_generators,
    )


@tracing.function
def run_partial_initial(device):
    from .common.initial import InitialConfig

    tracing_connector.get().set_device_attributes(tracing_connector.get().get_current_span(), device)

    run_args = GeneratorPartialRunArgs(device)
    return run_partial_generators([InitialConfig(storage=device.storage, do_run=True)], [], run_args)


@tracing.function
def run_partial_generators(
    gens: List["PartialGenerator"],
    ref_gens: List["RefGenerator"],
    run_args: GeneratorPartialRunArgs,
):
    logger = get_logger(host=run_args.device.hostname)
    tracing_connector.get().set_device_attributes(tracing_connector.get().get_current_span(), run_args.device)

    ret = RunGeneratorResult()
    if run_args.generators_context is not None:
        os.environ["ANN_GENERATORS_CONTEXT"] = run_args.generators_context

    for gen in ref_gens:
        ret.ref_matcher.add(gen.ref(run_args.device), gen)

    logger.debug("Generating selected PARTIALs ...")

    for gen in gens:
        try:
            result = _run_partial_generator(gen, run_args)
        except NotSupportedDevice as exc:
            logger.info("generator %s raised unsupported error: %r", gen, exc)
            continue

        if not result:
            continue

        ref_match = ret.ref_matcher.match(result.config)
        for ref_gen, groups in ref_match:
            gens.append(ref_gen.with_groups(groups))
            ret.ref_track.add(gen.__class__, ref_gen.__class__)

        ret.ref_track.config(gen.__class__, result.config)
        ret.add_partial(result)

    return ret


@tracing.function(name="run_partial_generator")
def _run_partial_generator(gen: "PartialGenerator", run_args: GeneratorPartialRunArgs) -> Optional[GeneratorPartialResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))
    device = run_args.device
    output = ""
    config = odict()
    safe_config = odict()

    if not gen.supports_device(device):
        logger.info("generator %s is not supported for device %s", gen, device.hostname)
        return None

    span = tracing_connector.get().get_current_span()
    if span:
        tracing_connector.get().set_device_attributes(span, run_args.device)
        tracing_connector.get().set_dimensions_attributes(span, gen, run_args.device)
        span.set_attributes({
            "use_acl": run_args.use_acl,
            "use_acl_safe": run_args.use_acl_safe,
            "generators_context": str(run_args.generators_context),
        })

    with GeneratorPerfMesurer(gen, run_args=run_args) as pm:
        if not run_args.no_new:
            if gen.get_user_runner(device):
                logger.info("Generating PARTIAL ...")
            try:
                output = gen(device, run_args.annotate)
            except NotSupportedDevice:
                # это исключение нужно передать выше в оригинальном виде
                raise
            except Exception as err:
                filename, lineno = gen.get_running_line()
                logger.exception("Generator error in file '%s:%i'", filename, lineno)
                raise GeneratorError(f"{gen} on {device}") from err

            fmtr = tabparser.make_formatter(device.hw)
            try:
                config = tabparser.parse_to_tree(text=output, splitter=fmtr.split)
            except tabparser.ParserError as err:
                logger.exception("Parser error")
                raise GeneratorError from err

    acl = gen.acl(device) or ""
    rules = compile_acl_text(textwrap.dedent(acl), device.hw.vendor)
    acl_safe = gen.acl_safe(device) or ""
    safe_rules = compile_acl_text(textwrap.dedent(acl_safe), device.hw.vendor)

    if run_args.use_acl:
        try:
            with tracing_connector.get().start_as_current_span("apply_acl", tracer_name=__name__, min_duration="0.01") as acl_span:
                tracing_connector.get().set_device_attributes(acl_span, run_args.device)
                config = patching.apply_acl(
                    config=config,
                    rules=rules,
                    fatal_acl=True,
                    with_annotations=run_args.annotate,
                )
            if run_args.use_acl_safe:
                with tracing_connector.get().start_as_current_span(
                    "apply_acl_safe",
                    tracer_name=__name__,
                    min_duration="0.01"
                ) as acl_safe_span:
                    tracing_connector.get().set_device_attributes(acl_safe_span, run_args.device)
                    safe_config = patching.apply_acl(
                        config=config,
                        rules=safe_rules,
                        fatal_acl=False,
                        with_annotations=run_args.annotate,
                    )
        except patching.AclError as err:
            logger.error("ACL error: generator is not allowed to yield this command: %s", err)
            raise GeneratorError from err
        except NotImplementedError as err:
            logger.error(str(err))
            raise GeneratorError from err

    return GeneratorPartialResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        output=output,
        acl=acl,
        acl_rules=rules,
        acl_safe=acl_safe,
        acl_safe_rules=safe_rules,
        config=config,
        safe_config=safe_config,
        perf=pm.last_result,
    )


@tracing.function
def check_entire_generators_required_packages(gens, device_packages: FrozenSet[str]) -> List[str]:
    errors: List[str] = []
    for gen in gens:
        if not gen.REQUIRED_PACKAGES.issubset(device_packages):
            missing = gen.REQUIRED_PACKAGES - device_packages
            missing_str = ", ".join("`{}'".format(pkg) for pkg in sorted(missing))
            if len(missing) == 1:
                errors.append("missing package {} required for {}".format(missing_str, gen))
            else:
                errors.append("missing packages {} required for {}".format(missing_str, gen))
    return errors


@tracing.function
def run_file_generators(
        gens: Iterable[Union["JSONFragment", "Entire"]],
        device: "Device",
) -> RunGeneratorResult:
    """Run generators that generate files or file parts."""
    ret = RunGeneratorResult()
    logger = get_logger(host=device.hostname)
    logger.debug("Generating selected ENTIREs and JSON_FRAGMENTs ...")
    for gen in gens:
        if gen.__class__.TYPE == "ENTIRE":
            run_generator_fn = _run_entire_generator
            add_result_fn = ret.add_entire
        elif gen.__class__.TYPE == "JSON_FRAGMENT":
            run_generator_fn = _run_json_fragment_generator
            add_result_fn = ret.add_json_fragment
        else:
            raise RuntimeError(f"Unknown generator class type: cls={gen.__class__} TYPE={gen.__class__.TYPE}")
        try:
            result = run_generator_fn(gen, device)
        except NotSupportedDevice as exc:
            logger.info("generator %s raised unsupported error: %r", gen, exc)
            continue
        if result:
            add_result_fn(result)

    return ret


@tracing.function(min_duration="0.5")
def _run_entire_generator(gen: "Entire", device: "Device") -> Optional[GeneratorResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))
    if not gen.supports_device(device):
        logger.info("generator %s is not supported for device %s", gen, device.hostname)
        return

    span = tracing_connector.get().get_current_span()
    if span:
        tracing_connector.get().set_device_attributes(span, device)
        tracing_connector.get().set_dimensions_attributes(span, gen, device)

    path = gen.path(device)
    if not path:
        raise RuntimeError("entire generator should return non-empty path")

    logger.info("Generating ENTIRE ...")
    with GeneratorPerfMesurer(gen, trace_min_duration="0.5") as pm:
        output = gen(device)

    return GeneratorEntireResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        path=path,
        output=output,
        reload=gen.get_reload_cmds(device),
        prio=gen.prio,
        perf=pm.last_result,
        is_safe=gen.is_safe(device),
    )


def _make_generator_ctx(gen):
    return "%s.[%s]" % (gen.__module__, gen.__class__.__name__)


def _run_json_fragment_generator(
        gen: "JSONFragment",
        device: "Device",
) -> Optional[GeneratorResult]:
    logger = get_logger(generator=_make_generator_ctx(gen))
    if not gen.supports_device(device):
        logger.info("generator %s is not supported for device %s", gen, device.hostname)
        return

    path = gen.path(device)
    if not path:
        raise RuntimeError("json fragment generator should return non-empty path")

    acl_item_or_list_of_items = gen.acl(device)
    safe_acl_item_or_list_of_items = gen.acl_safe(device)
    if not acl_item_or_list_of_items:
        raise RuntimeError("json fragment generator should return non-empty acl")
    if isinstance(acl_item_or_list_of_items, list):
        acl = acl_item_or_list_of_items
    else:
        acl = [acl_item_or_list_of_items]
    if isinstance(safe_acl_item_or_list_of_items, list):
        acl_safe = safe_acl_item_or_list_of_items
    else:
        acl_safe = [safe_acl_item_or_list_of_items]

    logger.info("Generating JSON_FRAGMENT ...")
    with GeneratorPerfMesurer(gen) as pm:
        config = gen(device)
    reload_cmds = gen.get_reload_cmds(device)
    return GeneratorJSONFragmentResult(
        name=gen.__class__.__name__,
        tags=gen.TAGS,
        path=path,
        acl=acl,
        acl_safe=acl_safe,
        config=config,
        reload=reload_cmds,
        perf=pm.last_result,
        reload_prio=gen.reload_prio,
    )


def _get_generators(module_paths: Union[List[str], dict], storage, device=None):
    if isinstance(module_paths, dict):
        if device is None:
            module_paths = module_paths.get("default")
        else:
            modules = []
            seen = set()
            matched_property = None
            for prop, prop_modules in module_paths.get("per_device_property", {}).items():
                if getattr(device, prop, False) is True:
                    if matched_property is not None:
                        raise RuntimeError(
                            f"Device {device.hostname} is matched by more than one "
                            f"per_device_property: {matched_property} and {prop}"
                        )
                    matched_property = prop
                    for module in prop_modules:
                        if module not in seen:
                            modules.append(module)
                            seen.add(module)
            module_paths = modules or module_paths.get("default")
    res_generators = []
    for module_path in module_paths:
        module = importlib.import_module(module_path)
        if hasattr(module, "get_generators"):
            generators: List[BaseGenerator] = module.get_generators(storage)
            res_generators += generators
    return res_generators


def _get_ref_generators(module_paths: List[str], storage, device):
    if isinstance(module_paths, dict):
        module_paths = module_paths.get("default")
    res_generators = []
    for module_path in module_paths:
        module = importlib.import_module(module_path)
        if hasattr(module, "get_ref_generators"):
            res_generators += module.get_ref_generators(storage)
    return res_generators


def select_generators(gens: GenSelectOptions, classes: Iterable[BaseGenerator]):
    def contains(obj, where):
        if where:
            return obj.get_aliases().intersection(where)
        return False

    def has(cls, what):
        return what in cls.TAGS

    flts = [lambda c: not isinstance(c, RefGenerator)]
    if gens.allowed_gens:
        flts.append(lambda c: contains(c, gens.allowed_gens))
    elif gens.force_enabled:
        flts.append(lambda c: not has(c, DISABLED_TAG) or contains(c, gens.force_enabled))
    elif not gens.ignore_disabled:
        flts.append(lambda c: not has(c, DISABLED_TAG))

    if gens.excluded_gens:
        flts.append(lambda c: not contains(c, gens.excluded_gens))

    return filter(lambda x: all(f(x) for f in flts), classes)



from typing import Dict, Any, Optional

from dataclass_rest.exceptions import ClientError

from annet.storage import StorageProvider, Storage
from annet.connectors import AdapterWithName, AdapterWithConfig, T
from .common.status_client import NetboxStatusClient
from .common.storage_opts import NetboxStorageOpts
from .common.query import NetboxQuery
from .v24.storage import NetboxStorageV24
from .v37.storage import NetboxStorageV37


def storage_factory(opts: NetboxStorageOpts) -> Storage:
    client = NetboxStatusClient(opts.url, opts.token)
    try:
        status = client.status()
    except ClientError as e:
        if e.status_code == 404:
            # old version do not support status reqeust
            return NetboxStorageV24(opts)
        raise
    if status.netbox_version.startswith("3."):
        return NetboxStorageV37(opts)
    else:
        raise ValueError(f"Unsupported version: {status.netbox_version}")


class NetboxProvider(StorageProvider, AdapterWithName, AdapterWithConfig):
    def __init__(self, url: Optional[str] = None, token: Optional[str] = None):
        self.url = url
        self.token = token

    @classmethod
    def with_config(cls, **kwargs: Dict[str, Any]) -> T:
        return cls(**kwargs)

    def storage(self):
        return storage_factory

    def opts(self):
        return NetboxStorageOpts

    def query(self):
        return NetboxQuery

    @classmethod
    def name(cls) -> str:
        return "netbox"



import copy
import operator
import textwrap
from collections import OrderedDict as odict
from typing import (  # pylint: disable=unused-import
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from .lib import jun_activate, merge_dicts, strip_annotation, uniq
from .rbparser import platform
from .rbparser.ordering import compile_ordering_text
from .rulebook.common import call_diff_logic
from .rulebook.common import default as common_default
from .tabparser import CommonFormatter
from .types import Diff, Op


# =====
class AclError(Exception):
    pass


class AclNotExclusiveError(AclError):
    pass


class PatchRow:
    row: str

    def __init__(self, row: str):
        self.row = row

    def __eq__(self, other: object) -> bool:
        if isinstance(other, str):
            return self.row == other
        if not isinstance(other, PatchRow):
            return NotImplemented
        return self.row == other.row

    def __hash__(self) -> int:
        return hash(self.row)

    def __str__(self) -> str:
        return self.row


class PatchItem:
    row: str
    child: "Union[PatchTree, None]"
    context: Dict[str, str]

    def __init__(self, row, child, context):
        self.row = row
        self.child = child
        self.context = context

    def __str__(self):
        return (
            f"PatchItem(\n"
            f'    row="{self.row}",\n'
            f"    child={textwrap.indent(str(self.child), '    ').strip()},\n"
            f"    context={self.context}\n"
            f")"
        )


class PatchTree:
    itms: List[PatchItem]

    def __init__(self, row: Optional[str] = None):
        self.itms = []
        if row:
            self.add(row, {})

    def add(self, row: str, context: Dict[str, str]) -> None:
        self.itms.append(PatchItem(row, None, context))

    def add_block(self, row: str, subtree: "Optional[PatchTree]" = None, context: Dict[str, str] = None) -> "PatchTree":
        if subtree is None:
            subtree = PatchTree()
        if context is None:
            context = {}
        self.itms.append(PatchItem(row, subtree, context))
        return subtree

    def items(self) -> "Iterator[Tuple[str, Union[PatchTree, None]]]":
        for item in self.itms:
            yield str(item.row), item.child

    def asdict(self) -> Dict:
        ret = odict()
        for row in uniq(i.row for i in self.itms):
            subtrees = []
            for i in self.itms:
                if i.row == row and i.child is not None:
                    subtrees.append(i.child.asdict())
            if subtrees:
                ret[str(row)] = merge_dicts(*subtrees)
            else:
                ret[str(row)] = None
        return ret

    def __bool__(self):
        return bool(self.itms)

    def __str__(self):
        n = ",\n"
        itms = map(lambda x: textwrap.indent(str(x), "    "), self.itms)
        return (
            f"PatchTree(\n"
            f"    itms=[\n"
            f"{textwrap.indent(n.join(itms), '    ')}\n"
            f"    ]\n"
            f")"
        )


class Orderer:
    def __init__(self, rb, vendor):
        self.rb = rb
        self.vendor = vendor

    def ref_insert(self, ref_tracker):
        for ref, _ in reversed(ref_tracker.configs()):
            self.insert(ref)
        for _, defs in reversed(ref_tracker.configs()):
            self.insert(defs)

    def insert(self, rules):
        if isinstance(rules, dict):
            fmtr = CommonFormatter()
            rules = fmtr.join(rules)
        rules = compile_ordering_text(rules, self.vendor)
        self.rb = merge_dicts(rules, self.rb)

    def rule_weight(self, row, rule, regexp_key):
        return len(set(row).intersection(set(rule["attrs"][regexp_key].pattern))) / len(row)

    def get_order(self, row, cmd_direct):
        f_order = None
        f_weight = 0
        f_rule = ""
        children = []
        ordering = self.rb
        block_exit = platform.VENDOR_EXIT[self.vendor]

        for (order, (raw_rule, rule)) in enumerate(ordering.items()):
            direct_matched = bool(rule["attrs"]["direct_regexp"].match(row))
            if not rule["attrs"]["order_reverse"] and (direct_matched or rule["attrs"]["reverse_regexp"].match(row)):
                # если не указано order_reverse - правило считается прямым
                regexp_key = ("direct_regexp" if direct_matched else "reverse_regexp")
                weight = self.rule_weight(row, rule, regexp_key)
                if f_order is None or f_weight < weight:
                    f_order = order
                    f_weight = weight
                    f_rule = (raw_rule, rule["attrs"][regexp_key])
                children.extend(ordering[raw_rule]["children"].items())

            elif rule["attrs"]["order_reverse"] and not cmd_direct and direct_matched:
                weight = self.rule_weight(row, rule, "direct_regexp")
                if f_order is None or f_weight < weight or (f_weight == weight and not cmd_direct):
                    f_order = order
                    f_weight = weight
                    f_rule = (raw_rule, rule["attrs"]["direct_regexp"])
                    cmd_direct = True
                children = []

            elif block_exit and block_exit == row:
                f_order = float("inf")
                f_rule = (raw_rule, block_exit)
                cmd_direct = True
                children = []

        return (f_order or 0), cmd_direct, odict(children), f_rule

    def order_config(self, config):
        if self.vendor not in platform.VENDOR_REVERSES:
            return config

        ordered = []
        reverse_prefix = platform.VENDOR_REVERSES[self.vendor]
        if not config:
            return odict()
        for (row, children) in config.items():
            cmd_direct = not row.startswith(reverse_prefix)
            (order, direct, rb, _) = self.get_order(row, cmd_direct)
            child_orderer = Orderer(rb, self.vendor)
            children = child_orderer.order_config(children)
            ordered.append({
                "row": row,
                "children": children,
                "direct": direct,
                "order": order,
            })

        return odict(
            (item["row"], item["children"])
            for item in sorted(ordered, key=(lambda item: (
                (item["order"] if item["direct"] else -item["order"]),
                item["direct"],
            )))
        )


# =====
def apply_acl(config, rules, fatal_acl=False, exclusive=False, with_annotations=False, _path=()):
    passed = odict()
    for (row, children) in config.items():
        if with_annotations:
            # do not pass annotations through ACL
            test_row = strip_annotation(row)
        else:
            test_row = row
        try:
            (match, children_rules) = match_row_to_acl(test_row, rules, exclusive)
        except AclNotExclusiveError as err:
            raise AclNotExclusiveError("'%s', %s" % ("/ ".join(_path + (row,)), err))
        if match:
            if not (match["is_reverse"] and all(match["attrs"]["cant_delete"])):
                passed[row] = apply_acl(
                    config=children,
                    rules=children_rules,
                    fatal_acl=fatal_acl,
                    exclusive=exclusive,
                    with_annotations=with_annotations,
                    _path=_path + (row,)
                )
        elif fatal_acl:
            raise AclError(" / ".join(_path + (row,)))
    return passed


def apply_acl_diff(diff, rules):
    passed = []
    for (op, row, children, d_match) in diff:
        (match, children_rules) = match_row_to_acl(row, rules)
        if match:
            if op == Op.REMOVED and all(match["attrs"]["cant_delete"]):
                op = Op.AFFECTED
            children = apply_acl_diff(children, children_rules)
            passed.append((op, row, children, d_match))
    return passed


def mark_unchanged(diff):
    passed = []
    for (op, row, children, d_match) in diff:
        if op == Op.AFFECTED:
            children = mark_unchanged(children)
            if all(x[0] == Op.UNCHANGED for x in children):
                op = Op.UNCHANGED
        passed.append((op, row, children, d_match))
    return passed


def strip_unchanged(diff):
    passed = []
    for (op, row, children, d_match) in diff:
        if op == Op.UNCHANGED:
            continue
        children = strip_unchanged(children)
        passed.append((op, row, children, d_match))
    return passed


def make_diff(old, new, rb, acl_rules_list) -> Diff:
    # не позволяем logic-коду модифицировать конфиг
    old = copy.deepcopy(old)
    new = copy.deepcopy(new)
    diff_pre = apply_diff_rb(old, new, rb)
    diff = call_diff_logic(diff_pre, old, new)
    for acl_rules in acl_rules_list:
        if acl_rules is not None:
            diff = apply_acl_diff(diff, acl_rules)
    diff = mark_unchanged(diff)
    return diff


def apply_diff_rb(old, new, rb):
    """ Diff pre is a odict {(key, diff_logic): {}} """
    diff_pre = odict()
    for row in list(uniq(old, new)):
        (match, children_rules) = _match_row_to_rules(row, rb["patching"])
        if match:
            diff_pre[row] = {
                "match": match,
                "subtree": apply_diff_rb(
                    old.get(row, odict()),
                    new.get(row, odict()),
                    rb={"patching": children_rules},  # Нужен только кусок, касающийся правил для патчей
                ),
            }
        else:
            old.pop(row, None)
            new.pop(row, None)
    return diff_pre


def make_pre(diff: Diff, _parent_match=None) -> Dict[str, Any]:
    pre = odict()
    for (op, row, children, match) in diff:
        if _parent_match and _parent_match["attrs"]["multiline"]:
            # Если родительское правило было мультилайном, то все внутренности станут его контентом.
            # Это значит, что к ним будет принудительно применяться common.default() и фейковое
            # правило __MULTILINE_BODY__.
            match = {
                "raw_rule": "__MULTILINE_BODY__",
                "key": row,
                "attrs": {
                    "comment": [],
                    "logic": common_default,  # Прекрасно работает с мультилайнами и обрезанным правилом
                    "multiline": True,
                    "context": _parent_match["attrs"]["context"],
                }
            }
        raw_rule = match["raw_rule"]
        key = match["key"]

        if raw_rule not in pre:
            pre[raw_rule] = {
                "attrs": match["attrs"],
                "items": odict(),
            }
        if key not in pre[raw_rule]["items"]:
            pre[raw_rule]["items"][key] = {
                Op.ADDED: [],
                Op.REMOVED: [],
                Op.MOVED: [],
                Op.AFFECTED: [],
                Op.UNCHANGED: [],
            }

        pre[raw_rule]["items"][key][op].append({
            "row": row,
            "children": make_pre(
                diff=children,
                _parent_match=match,
            ),
        })
    return pre


_comment_macros = {
    "!!HYES!!": "!!question!![Y/N]!!answer!!Y!! !!question!![y/n]!!answer!!Y!! !!question!![Yes/All/No/Cancel]!!answer!!Y!!"
}


def make_patch(pre, rb, hw, add_comments, orderer=None, _root_pre=None, do_commit=True):
    patch = []
    if not orderer:
        orderer = Orderer(rb["ordering"], hw.vendor)

    for (raw_rule, content) in pre.items():
        for (key, diff) in content["items"].items():
            # чтобы logic не мог поменять атрибуты
            rule_pre = content.copy()
            attrs = rule_pre["attrs"].copy()

            iterable = attrs["logic"](
                rule=attrs,
                key=key,
                diff=diff,
                hw=hw,
                rule_pre=rule_pre,
                root_pre=(_root_pre or pre),
            )
            for (direct, row, sub_pre) in iterable:
                if direct is not None:
                    patch_row = row
                    if add_comments:
                        comments = " ".join(attrs["comment"])
                        for (macro, m_value) in _comment_macros.items():
                            comments = comments.replace(macro, m_value)
                        if comments:
                            patch_row = "%s %s" % (row, comments)

                    # pylint: disable=unused-variable
                    (order, order_direct, ordering, order_rule) = orderer.get_order(row, direct)
                    fmt_row = patch_row
                    # fmt_row += "  # %s" % str(order_rule)  # uncomment to debug ordering

                    if not do_commit and attrs.get("force_commit", False):
                        # if do_commit is false skip patch that couldn't be applied without commit
                        continue

                    patch.append({
                        "row": fmt_row,
                        "children": (PatchTree() if not sub_pre else make_patch(
                            pre=sub_pre,
                            rb={"ordering": ordering},  # Нужен только кусок, касающийся правил для ордеринга
                            hw=hw,
                            add_comments=add_comments,
                            _root_pre=(_root_pre or pre),
                            do_commit=do_commit,
                        )),
                        "raw_rule": raw_rule,
                        "direct": direct,
                        "order": order,
                        "order_direct": order_direct,
                        "parent": attrs.get("parent", False),
                        "force_commit": attrs.get("force_commit", False),
                        "ignore_case": attrs.get("ignore_case", False),
                        "context": attrs["context"],
                    })
    tree = PatchTree()
    sorted_patch = sorted(patch, key=(lambda item: (
        (item["order"] if item["order_direct"] else -item["order"]),
        item["raw_rule"],
        item["order_direct"],
    )))
    for item in sorted_patch:
        if (not item["children"] and not item["parent"]) or not item["direct"]:
            tree.add(item["row"], item["context"])
        else:
            tree.add_block(item["row"], item["children"], item["context"])
        if item["force_commit"]:
            tree.add("commit", item["context"])
    return tree


def match_row_to_acl(row, rules, exclusive=False):
    matches = _find_acl_matches(row, rules)
    if matches:
        if exclusive:
            gen_cant_delete = {}
            for match in matches:
                names = match[0][0]["attrs"]["generator_names"]
                flags = match[0][0]["attrs"]["cant_delete"]
                for name, flag in zip(names, flags):
                    if name not in gen_cant_delete:
                        gen_cant_delete[name] = flag
                    else:
                        gen_cant_delete[name] &= flag
            can_delete = {name: flag for name, flag in gen_cant_delete.items() if not flag}
            if len(can_delete) > 1:
                generator_names = ", ".join(can_delete.keys())
                raise AclNotExclusiveError("generators: '%s'" % generator_names)
        return _select_match(matches, rules)
    return (None, None)  # (match, children_rules)


def _match_row_to_rules(row, rules):
    matches = _find_rules_matches(row, rules)
    if matches:
        return _select_match(matches, rules)
    return (None, None)


def _find_acl_matches(row, rules):
    res = []
    for regexp_key in ["direct_regexp", "reverse_regexp"]:
        for ((_, rule), is_global) in _rules_local_global(rules):
            row_to_match = _normalize_row_for_acl(row, rule)
            match = rule["attrs"][regexp_key].match(row_to_match)
            if match:
                rule["attrs"]["match"] = match.groupdict()
                # FIXME: сейчас у нас вообще не используется тип ignore, но он иногда встречается в ACL.
                # Проблема в том, что ACL мержится, и игноры все ломают. Надо придумать, что с этим сделать.
                # В данный момент ignore acl работает только в filter-acl, так как он целостный и накладывается независимо
                # В этом случае ignore правила так же матчатся и считается их специфичность на ряду с normal
                # при выборе ignore правила, заматченная строка не будет пропущена
                metric = (
                    rule["attrs"]["prio"],
                    # Calculate how specific matched regex is for the row
                    # based on how many symbols they share
                    len(set(row).intersection(set(rule["attrs"][regexp_key].pattern))) / len(row),
                )
                item = (
                    metric,
                    ((rule, (not is_global and regexp_key == "direct_regexp" and rule["type"] != "ignore")),
                     #       ^^^ is_cr_allowed ^^^    cr == children rules
                     {"is_reverse": (regexp_key == "reverse_regexp")}),
                    # ^^^ is_reverse ^^^
                )
                res.append(item)
    res.sort(key=operator.itemgetter(0), reverse=True)
    return [item[1] for item in res]


def _find_rules_matches(row, rules):
    matches = []
    for ((raw_rule, rule), is_global) in _rules_local_global(rules):
        match = rule["attrs"]["regexp"].match(row)
        if match:
            if rule["type"] == "ignore":
                return []
            matches.append(((rule, (not is_global)), {"raw_rule": raw_rule, "key": match.groups()}))
            #                       ^^^ is_cr_allowed
    return matches


def _select_match(matches, rules):
    ((f_rule, is_f_cr_allowed), f_other) = matches[0]  # f == first
    if f_rule["type"] == "ignore":
        # В данный момент эта ветка достижима только в filter-acl
        return (None, None)

    # Мерджим всех потомков которые заматчились
    local_children = odict()
    global_children = odict()
    if is_f_cr_allowed:
        for (rule, is_cr_allowed) in map(operator.itemgetter(0), matches):
            if is_cr_allowed:
                local_children = merge_dicts(local_children, rule["children"]["local"])
            # optional break on is_cr_allowed==False?

                global_children = merge_dicts(global_children, rule["children"]["global"])

    global_children = merge_dicts(global_children, rules["global"])

    children_rules = {
        "local": local_children,
        "global": global_children,
    }

    match = {"attrs": f_rule["attrs"]}
    match.update(f_other)
    return (match, children_rules)


def _rules_local_global(rules):
    for (raw_rule, rule) in rules["local"].items():
        yield ((raw_rule, rule), False)
    for (raw_rule, rule) in rules["global"].items():
        yield ((raw_rule, rule), True)


def _normalize_row_for_acl(row, rule):
    # NOCDEV-5940 У джуниперов есть служебрая разметка "inactive:"
    if rule["attrs"]["vendor"] == "juniper":
        row = jun_activate(row)
    return row



import asyncio
import logging
import multiprocessing
import os
import platform
import resource
import signal
import statistics
import time
from abc import ABC, abstractmethod
from functools import partial
from operator import itemgetter
from queue import Empty
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import colorama
import psutil

import annet.lib
from annet.annlib.command import Command, CommandList, Question  # noqa: F401
from annet.storage import Device


_logger = logging.getLogger(__name__)
FIRST_EXCEPTION = 1
ALL_COMPLETED = 2


class CommandResult(ABC):
    @abstractmethod
    def get_out(self) -> str:
        pass


class Connector(ABC):
    @abstractmethod
    async def cmd(self, cmd: Union[Command, str]) -> CommandResult:
        pass

    @abstractmethod
    async def download(self, files: List[str]) -> Dict[str, str]:
        pass

    @abstractmethod
    async def upload(self, files: Dict[str, str]):
        pass

    @abstractmethod
    def get_conn_trace(self) -> str:
        pass

    @abstractmethod
    async def aclose(self) -> str:
        pass


class Executor(ABC):
    # method for bulk config downloading TODO: remove in favor Connector.cmd
    @abstractmethod
    def fetch(self,
              devices: List[Device],
              files_to_download: Dict[str, List[str]] = None) -> Tuple[Dict[Device, str], Dict[Device, Any]]:
        pass

    @abstractmethod
    async def amake_connection(self, device: Device) -> Connector:
        pass


class ExecutorException(Exception):
    def __init__(self, *args: List[Any], auxiliary: Optional[Any] = None, **kwargs: object):
        self.auxiliary = auxiliary
        super().__init__(*args, **kwargs)

    def __repr__(self) -> str:
        return "%s(args=%r,auxiliary=%s)" % (self.__class__.__name__, self.args, self.auxiliary)


class ExecException(ExecutorException):
    def __init__(self, msg: str, cmd: str, res: str, **kwargs):
        super().__init__(**kwargs)
        self.args = msg, cmd, res
        self.kwargs = kwargs
        self.msg = msg
        self.cmd = cmd
        self.res = res

    def __str__(self) -> str:
        return str(self.msg)

    def __repr__(self) -> str:
        return "%s<%s, %s>" % (self.__class__.__name__, self.msg, self.cmd)


class BadCommand(ExecException):
    pass


class NonzeroRetcode(ExecException):
    pass


class CommitException(ExecException):
    pass

def chunks_tuple(l, n):  # noqa
    return [tuple(l[i:i + n]) for i in range(0, len(l), n)]


def async_bulk(
        executor: Executor,
        devices: List[Device],
        coro_gen: Callable[[Connector, Device], Any],
        *args,
        processes: int = 1,
        show_report: bool = True,
        do_log: bool = True,
        log_dir: Optional[str] = None,
        policy: int = ALL_COMPLETED,
        **kwargs,
):
    """Connect to specified devices and work with their CLI.

    Note: this function is not allowed to be run in parallel, since it's using global state (TODO: fixme)

    :param devices: List of devices' fqdns to use their CLI.
    :param coro_gen: Async function. It contains all logic about usage of CLI.
        See docstring of "bind_coro_args" for allowed function signature and examples.
    :param args: Positional arguments to "bulk" function.
    :type processes: Amount of processes to fork for current work.
    :param show_report: Set this flag to show report to stdout.
    :param do_log: If True and log_dir is not set, then log_dir will be filled automatically.
    :param log_dir: Specify path to log all response from devices.
    :param policy: int flag. If FIRST_EXCEPTION, then work will be stopped after first error.
        Otherwise all hosts will be processed.
        TODO: fix that policy is not used if processes=1
    :param kwargs: other arguments to pass to "bulk" function.
        Note: it is not passed directly to "coro_gen" function!
            kwargs should be {'kwargs': {'var1': value1}} to set "var1" with "value1" in "coro_gen" function.

    TODOs:
        * do not log if do_log=False and log_dir is set.

    """
    res = {}
    deploy_durations = {}
    kwargs["log_dir"] = log_dir
    kwargs["policy"] = policy

    if processes == 1:
        host_res, host_duration = annet.lib.do_async(bulk(executor, devices, coro_gen, *args, **kwargs))
        res.update(host_res)
        deploy_durations.update(host_duration)
    else:
        # FIXME: show_report works per process
        if len(devices) != len(set(devices)):
            raise Exception("hostnames should be unique")
        # warm up a cache
        # asyncio.get_event_loop().run_until_complete(get_validator_rt_data(hostnames))
        if isinstance(devices, dict):
            devices = list(devices.keys())
        hostnames_chunks = chunks_tuple(devices, int(len(devices) / processes) + 1)
        pool = {}
        for hostnames_chunk in hostnames_chunks:
            res_q = multiprocessing.Queue()
            p = multiprocessing.Process(target=_mp_async_bulk, args=[res_q, hostnames_chunk, coro_gen, *args], kwargs=kwargs)
            pool[p] = [res_q, hostnames_chunk]
            p.start()
            _logger.info("process (id=%d) work with %d chunks", p.pid, len(hostnames_chunks))

        seen_error = False
        while True:
            done = []
            for p in pool:
                host_res = None
                try:
                    # proc wont be exited till q.get() call
                    host_res, host_duration = pool[p][0].get(timeout=0.2)
                except Empty:
                    pass
                else:
                    done.append(p)

                if not p.is_alive() and not host_res:
                    _logger.error("process %s has died: hostnames: %s", p.pid, pool[p][1])
                    host_res = {hostname: Exception("died with exitcode %s" % p.exitcode) for hostname in pool[p][1]}
                    host_duration = {hostname: 0 for hostname in pool[p][1]}  # FIXME:
                    done.append(p)

                if host_res:
                    res.update(host_res)
                    deploy_durations.update(host_duration)

                if p.exitcode:
                    _logger.error("process %s finished with bad exitcode %s", p.pid, p.exitcode)
                    seen_error = True
            for p in done:
                pool.pop(p)
            if policy == FIRST_EXCEPTION and seen_error:
                for p in pool:
                    p.terminate()
                    if p.is_alive():
                        time.sleep(0.4)
                    if p.is_alive():
                        os.kill(p.pid, signal.SIGKILL)
                    for hostname in pool[p][1]:
                        res[hostname] = Exception("force kill with exitcode %s" % p.exitcode)
                        deploy_durations[hostname] = 0  # FIXME:
            if not pool:
                break

    if show_report:
        show_bulk_report(devices, res, deploy_durations, do_log and log_dir)

    return res


def _show_type_summary(caption, items, total, stat_items=None):
    if items:
        if not stat_items:
            stat = ""
        else:
            avg = statistics.mean(stat_items)
            stat = "   %(min).1f/%(max).1f/%(avg).1f/%(stdev)s  (min/max/avg/stdev)" % dict(
                min=min(stat_items),
                max=max(stat_items),
                avg=avg,
                stdev="-" if len(stat_items) < 2 else "%.1f" % statistics.stdev(stat_items, xbar=avg)
            )

        print("%-8s %d of %d%s" % (caption, len(items), total, stat))


def show_bulk_report(hostnames, res, durations, log_dir):
    total = len(hostnames)
    if not total:
        return

    colorama.init()

    print("\n====== bulk deploy report ======")

    done = [host for (host, hres) in res.items() if not isinstance(hres, Exception)]
    cancelled = [host for (host, hres) in res.items() if isinstance(hres, asyncio.CancelledError)]
    failed = [host for (host, hres) in res.items() if isinstance(hres, Exception) and host not in cancelled]
    lost = [host for host in hostnames if host not in res]
    limit = 30

    _show_type_summary("Done :", done, total, [durations[h] for h in done])
    _print_limit(done, partial(_print_hostname, style=colorama.Fore.GREEN), limit, total)

    _show_type_summary("Failed :", failed, total, [durations[h] for h in failed])

    _print_limit(failed, partial(_print_failed, res=res), limit, total)

    _show_type_summary("Cancelled :", cancelled, total, [durations[h] for h in cancelled if durations[h] is not None])
    _print_limit(cancelled, partial(_print_hostname, style=colorama.Fore.RED), limit, total)

    _show_type_summary("Lost :", lost, total)
    _print_limit(lost, _print_hostname, limit, total)

    err_limit = 5
    if failed:
        errs = {}
        for hostname in failed:
            fmt_err = _format_exc(res[hostname])
            if fmt_err in errs:
                errs[fmt_err] += 1
            else:
                errs[fmt_err] = 1
        print("Top errors :")
        for fmt_err, n in sorted(errs.items(), key=itemgetter(1), reverse=True)[:err_limit]:
            print("  %-4d %s" % (n, fmt_err))
        print("\n", end="")

    if log_dir:
        print("See deploy logs in %s/\n" % os.path.relpath(log_dir))


def _format_exc(exc):
    if isinstance(exc, ExecException):
        cmd = str(exc.cmd)
        if len(cmd) > 50:
            cmd = cmd[:50] + "~.."
        return "'%s', cmd '%s'" % (exc.msg, cmd)
    elif isinstance(exc, ExecutorException):
        return "%s%r" % (exc.__class__.__name__, exc.args)  # исключить многословный auxiliary
    else:
        return repr(exc)


def _print_hostname(host, style=None):
    if style:
        host = style + host + colorama.Style.RESET_ALL
    print("  %s" % host)


def _print_limit(items, printer, limit, total, end="\n"):
    if not items:
        return
    if len(items) > limit and len(items) > total * 0.7:
        print("  ... %d hosts" % len(items))
    for host in items[:limit]:
        printer(host)
    if len(items) > limit:
        print("  ... %d more hosts" % (len(items) - limit))

    print(end, end="")


def _print_failed(host, res):
    exc = res[host]
    color = colorama.Fore.YELLOW if isinstance(exc, Warning) else colorama.Fore.RED
    print("  %s - %s" % (color + host + colorama.Style.RESET_ALL, _format_exc(exc)))


def _mp_async_bulk(res_q: multiprocessing.Queue, *args, **kwargs):
    res = annet.lib.do_async(bulk(*args, **kwargs))
    res_q.put(res)
    res_q.close()


async def bulk(
        executor: Executor,
        devices: List[Device],
        coro_gen: Callable[[Connector, Device, Optional[Dict[str, Any]]], Any],
        max_parallel: float = 100,
        policy: int = ALL_COMPLETED,
        log_dir: str = True,  # pylint: disable=unused-argument
        kwargs: Optional[dict] = None,
        console_log: bool = True
) -> Tuple[Dict[str, Any], Dict[str, float]]:
    """Connect to specified devices and work with their CLI.

    :param hostnames: List of devices' fqdns to use their CLI.
    :param coro_gen: Async function. It contains all logic about usage of CLI.
        See docstring of "bind_coro_args" for allowed function signature and examples.
    :param max_parallel: Upper border to CPU usage (in percentage 1 CPU = 100).
        If cpu usage is over, then tasks are trottled.
    :param policy: Flag to specify when tasks are completed.
    :param log_dir: Specify path to log all response from devices.
        TODO: fix default value.
    :param kwargs: Device independent arguments to call function. See @bind_coro_args for details.
    :param get_device: See "make_connection" for better understanding.
    :param device_cls: See "make_connection" for better understanding.
    :param streamer_cls: See "make_connection" for better understanding.
    :param console_log: If True and there is no handlers for root logger, then stderr will be used for logging.
    :return: two dicts with results per host and execution duration per host.

    """
    if console_log:
        init_log()

    tasks = []
    res = {}
    pending = set()
    tasks_to_device = {}
    time_of_start = {}
    deploy_durations = {}
    now = None
    if not kwargs:
        kwargs = {}

    def start_hook(device: Device):
        time_of_start[device.hostname] = time.monotonic()

    def end_hook(device: Device, task: asyncio.Task):
        duration = now - time_of_start[device.hostname]
        deploy_durations[device.hostname] = duration

        coro_exc = task.exception()
        if coro_exc:
            if policy == FIRST_EXCEPTION:
                _logger.error("%s %r", device.hostname, coro_exc, exc_info=coro_exc)
                _logger.info("Terminating all running tasks according to FIRST_EXCEPTION policy")
                res[device.hostname] = coro_exc
                raise CancelAllTasks
            else:
                if isinstance(coro_exc, AssertionError):
                    _logger.error("%s %r", device.hostname, coro_exc, exc_info=coro_exc)
                else:
                    _logger.error("%s %r", device.hostname, coro_exc)
            return coro_exc
        else:
            _logger.info("Finished in %0.2fs, hostname=%s", duration, device.hostname)
            return task.result()

    for device in devices:
        try:
            conn = await executor.amake_connection(device=device)
        except Exception as exc:
            _logger.error("failed to connect to %s %r", device.hostname, exc)
            res[device] = exc
            continue
        start_hook(device)
        task = asyncio.create_task(coro_gen(conn=conn, device=device, **kwargs))
        tasks_to_device[task] = device
        tasks.append(task)
    try:
        ndone = 0
        with CpuThrottler(asyncio.get_event_loop(), maximum=max_parallel) as throttler:
            while pending or tasks:
                left_fds = int(fd_left() / 2)  # better to be safe than sorry

                for _ in range(min(throttler.curr - len(pending), len(tasks), left_fds)):
                    pending.add(tasks.pop(0))
                if len(pending) == 0:
                    _logger.debug("empty pending list. tasks=%s throttler curr=%s left_fds=%s. waiting", len(tasks),
                                  throttler.curr, left_fds)
                    await asyncio.sleep(1)
                    continue
                example_host = next(iter(pending))
                _logger.debug("tasks status: %d pending, %d queued, pending example %s", len(pending), len(tasks), example_host)
                done, pending = await asyncio.wait(pending, return_when=asyncio.FIRST_COMPLETED)

                now = time.monotonic()
                for task in done:
                    device = tasks_to_device[task]
                    res[device] = end_hook(device, task)
                    ndone += 1
    except CancelAllTasks:
        exc = asyncio.CancelledError()

        now = time.monotonic()
        for device, task in _get_remaining(tasks, pending, tasks_to_device):
            res[device] = exc

            if device.hostname in time_of_start:
                duration = now - time_of_start[device.hostname]
            else:
                duration = None
            deploy_durations[device.hostname] = duration

            if not asyncio.iscoroutine(task):
                _logger.info("task %s", task)
                task.cancel()

    return res, deploy_durations


def init_log():
    streamer = logging.StreamHandler()
    fmt = logging.Formatter("%(asctime)s - %(filename)s:%(lineno)d - %(funcName)s() - %(levelname)s - %(message)s",
                            "%Y-%m-%d %H:%M:%S")
    streamer.setFormatter(fmt)
    if not logging.root.handlers:
        logging.root.addHandler(streamer)


class DeferredFileWrite:
    def __init__(self, file, mode="r"):
        self._file = file
        wrapper = {"w": "a", "wb": "ab"}
        if mode in wrapper:
            self._mode = wrapper[mode]
        else:
            raise Exception()

    def write(self, data):
        with open(self._file, self._mode) as fh:
            fh.write(data)

    def close(self):
        pass

    def flush(self):
        pass


class CancelAllTasks(Exception):
    pass


def _get_remaining(tasks, pending, tasks_to_device):
    for task in pending:
        yield (tasks_to_device[task], task)
    for task in tasks:
        yield (tasks_to_device[task], task)


_platform = platform.system()
_fd_available = resource.getrlimit(resource.RLIMIT_NOFILE)[0]


def fd_left():
    res = _fd_available
    if _platform == "Linux":
        res = _fd_available - len(list(os.scandir(path="/proc/self/fd/")))
    return res


class CpuThrottler:
    def __init__(self, loop, start=20, maximum=None, minimum=5, hz=1.0, target=80.0):
        self.loop = loop
        self.minimum = int(minimum)
        self.maximum = int(maximum or 0)
        self.hz = hz
        self.target = target
        self.timer_handle = None
        self.last_usage = 0
        self.curr = int(start)
        self.proc = psutil.Process(os.getpid())

    def __enter__(self):
        if self.proc and self.maximum:
            self.proc.cpu_percent()  # initialize previous value
            self.timer_handle = self.loop.call_later(self.hz, self.schedule)
        return self

    def __exit__(self, type_, value, tb):
        if self.timer_handle:
            self.timer_handle.cancel()

    def schedule(self):
        # re-schedule
        self.timer_handle = self.loop.call_later(self.hz, self.schedule)

        cpu_usage = self.proc.cpu_percent()
        self.last_usage = cpu_usage
        _logger.debug("current cpu_usage=%s", cpu_usage)
        if cpu_usage > self.target:
            self.change_by(0.5)
        elif cpu_usage > self.target * 0.8:
            pass
        elif cpu_usage > self.target * 0.2:
            self.change_by(1.2)
        else:
            self.change_by(1.5)

    def change_by(self, rate):
        new_curr = int(self.curr * rate)
        # округлим в нужную сторону
        if new_curr == self.curr:
            if rate > 1:
                new_curr += 1
            elif rate < 1:
                new_curr -= 1
        # ограничим пределами
        if new_curr < self.curr:
            new_curr = max(self.minimum, new_curr)
        else:
            if self.maximum is not None:
                new_curr = min(self.maximum, new_curr)

        if new_curr < self.curr:
            _logger.info("decreasing max_slots %d -> %d, cpu_usage=%.1f", self.curr, new_curr, self.last_usage)
        elif new_curr > self.curr:
            _logger.info("increasing max_slots %d -> %d, cpu_usage=%.1f", self.curr, new_curr, self.last_usage)

        # new_curr не делаем меньше 0, иначе не сможем его увеличить
        self.curr = max(new_curr, 1)



from dataclasses import dataclass
from typing import List, Union, Iterable, Optional

from annet.storage import Query


@dataclass
class NetboxQuery(Query):
    query: List[str]

    @classmethod
    def new(
            cls, query: Union[str, Iterable[str]],
            hosts_range: Optional[slice] = None,
    ) -> "NetboxQuery":
        if hosts_range is not None:
            raise ValueError("host_range is not supported")
        return cls(query=list(query))

    @property
    def globs(self):
        # We process every query host as a glob
        return self.query

    def is_empty(self) -> bool:
        return len(self.query) == 0



from annet.annlib.rulebook import common
from annet.annlib.types import Op


def syslog_level(rule, key, diff, **_):
    # syslog-level can be overwritten only
    if diff[Op.ADDED]:
        yield from common.default(rule, key, diff)
